import os

import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids
from autolens.data.array import scaled_array
from autolens.data import ccd
from autolens.data import simulated_ccd
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import ray_tracing

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)
test_positions_dir = "{}/../test_files/positions/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestSimulateCCD(object):
    def test__setup_image__correct_attributes(self):

        array = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])

        psf = ccd.PSF(array=3.0 * np.ones((3, 3)), pixel_scale=1.0)
        noise_map = 5.0 * np.ones((3, 3))

        ccd_data = simulated_ccd.SimulatedCCDData(
            image=array,
            pixel_scale=0.1,
            noise_map=noise_map,
            psf=psf,
            background_noise_map=7.0 * np.ones((3, 3)),
            poisson_noise_map=9.0 * np.ones((3, 3)),
            exposure_time_map=11.0 * np.ones((3, 3)),
        )

        assert ccd_data.image == pytest.approx(
            np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]), 1e-2
        )
        assert (ccd_data.psf == 3.0 * np.ones((3, 3))).all()
        assert (ccd_data.noise_map == 5.0 * np.ones((3, 3))).all()
        assert (ccd_data.background_noise_map == 7.0 * np.ones((3, 3))).all()
        assert (ccd_data.poisson_noise_map == 9.0 * np.ones((3, 3))).all()
        assert (ccd_data.exposure_time_map == 11.0 * np.ones((3, 3))).all()
        assert ccd_data.origin == (0.0, 0.0)

    def test__setup_with_all_features_off(self):

        image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            pixel_scale=0.1,
            add_noise=False,
        )

        assert (ccd_data_simulated.exposure_time_map == np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1
        assert (
            ccd_data_simulated.image
            == np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
        ).all()
        assert ccd_data_simulated.origin == (0.0, 0.0)

    def test__setup_with_background_sky_on__noise_off__no_noise_in_image__noise_map_is_noise_value(
        self
    ):
        image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=16.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            add_noise=False,
            noise_if_add_noise_false=0.2,
            noise_seed=1,
        )

        assert (ccd_data_simulated.exposure_time_map == 1.0 * np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1

        assert (
            ccd_data_simulated.image
            == np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
        ).all()
        assert ccd_data_simulated.noise_map == 0.2 * np.ones((3, 3))

        assert (ccd_data_simulated.background_noise_map == 4.0 * np.ones((3, 3))).all()

    def test__setup_with_background_sky_on__noise_on_so_background_adds_noise_to_image(
        self
    ):
        image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=16.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            add_noise=True,
            noise_seed=1,
        )

        assert (ccd_data_simulated.exposure_time_map == 1.0 * np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1

        assert (
            ccd_data_simulated.image
            == np.array([[1.0, 5.0, 4.0], [1.0, 2.0, 1.0], [5.0, 2.0, 7.0]])
        ).all()

        assert (
            ccd_data_simulated.poisson_noise_map
            == np.array(
                [
                    [np.sqrt(1.0), np.sqrt(5.0), np.sqrt(4.0)],
                    [np.sqrt(1.0), np.sqrt(2.0), np.sqrt(1.0)],
                    [np.sqrt(5.0), np.sqrt(2.0), np.sqrt(7.0)],
                ]
            )
        ).all()

        assert (ccd_data_simulated.background_noise_map == 4.0 * np.ones((3, 3))).all()

    def test__setup_with_psf_blurring_on__blurs_image_and_trims_psf_edge_off(self):
        image = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        psf = ccd.PSF(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scale=1.0,
        )

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            psf=psf,
            add_noise=False,
        )

        assert (
            ccd_data_simulated.image
            == np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]])
        ).all()
        assert (ccd_data_simulated.exposure_time_map == np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1

    def test__setup_with_background_sky_and_psf_on__psf_does_no_blurring__image_and_sky_both_trimmed(
        self
    ):
        image = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        psf = ccd.PSF(
            array=np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]),
            pixel_scale=1.0,
        )

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=16.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            psf=psf,
            background_sky_map=background_sky_map,
            add_noise=False,
            noise_seed=1,
        )

        assert (ccd_data_simulated.exposure_time_map == 1.0 * np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1

        assert (
            ccd_data_simulated.image
            == np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])
        ).all()

        assert (ccd_data_simulated.background_noise_map == 4.0 * np.ones((3, 3))).all()

    def test__setup_with_noise(self):
        image = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=20.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=20.0,
            exposure_time_map=exposure_time_map,
            add_noise=True,
            noise_seed=1,
        )

        assert (ccd_data_simulated.exposure_time_map == 20.0 * np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1

        assert ccd_data_simulated.image == pytest.approx(
            np.array([[0.0, 0.0, 0.0], [0.0, 1.05, 0.0], [0.0, 0.0, 0.0]]), 1e-2
        )

        # Because of the regular value is 1.05, the estimated Poisson noise_map_1d is:
        # sqrt((1.05 * 20))/20 = 0.2291

        assert ccd_data_simulated.poisson_noise_map == pytest.approx(
            np.array([[0.0, 0.0, 0.0], [0.0, 0.2291, 0.0], [0.0, 0.0, 0.0]]), 1e-2
        )

        assert ccd_data_simulated.noise_map == pytest.approx(
            np.array([[0.0, 0.0, 0.0], [0.0, 0.2291, 0.0], [0.0, 0.0, 0.0]]), 1e-2
        )

    def test__setup_with__psf_blurring_and_poisson_noise_on__poisson_noise_added_to_blurred_image(
        self
    ):

        image = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        psf = ccd.PSF(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scale=1.0,
        )

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=20.0, pixel_scale=0.1, shape=image.shape
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=20.0,
            exposure_time_map=exposure_time_map,
            psf=psf,
            add_noise=True,
            noise_seed=1,
        )

        assert (ccd_data_simulated.exposure_time_map == 20.0 * np.ones((3, 3))).all()
        assert ccd_data_simulated.pixel_scale == 0.1
        assert ccd_data_simulated.image == pytest.approx(
            np.array([[0.0, 1.05, 0.0], [1.3, 2.35, 1.05], [0.0, 1.05, 0.0]]), 1e-2
        )

        # The estimated Poisson noises are:
        # sqrt((2.35 * 20))/20 = 0.3427
        # sqrt((1.3 * 20))/20 = 0.2549
        # sqrt((1.05 * 20))/20 = 0.2291

        assert ccd_data_simulated.poisson_noise_map == pytest.approx(
            np.array(
                [[0.0, 0.2291, 0.0], [0.2549, 0.3427, 0.2291], [0.0, 0.2291, 0.0]]
            ),
            1e-2,
        )

    def test__simulate_function__turns_exposure_time_and_sky_level_to_arrays(self):

        image = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        psf = ccd.PSF(
            array=np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]]),
            pixel_scale=1.0,
        )

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=16.0, pixel_scale=0.1, shape=image.shape
        )

        simulated_ccd_variable = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            exposure_time=1.0,
            exposure_time_map=exposure_time_map,
            psf=psf,
            background_sky_map=background_sky_map,
            pixel_scale=0.1,
            add_noise=False,
            noise_seed=1,
        )

        image = np.array(
            [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=image,
            pixel_scale=0.1,
            exposure_time=1.0,
            background_sky_level=16.0,
            psf=psf,
            add_noise=False,
            noise_seed=1,
        )

        assert (
            simulated_ccd_variable.exposure_time_map
            == ccd_data_simulated.exposure_time_map
        ).all()
        assert simulated_ccd_variable.pixel_scale == ccd_data_simulated.pixel_scale
        assert simulated_ccd_variable.image == pytest.approx(
            ccd_data_simulated.image, 1e-4
        )
        assert (
            simulated_ccd_variable.background_noise_map
            == ccd_data_simulated.background_noise_map
        ).all()

    def test__noise_map_creates_nans_due_to_low_exposure_time__raises_error(self):

        image = np.ones((9, 9))

        psf = ccd.PSF.from_gaussian(shape=(3, 3), sigma=0.1, pixel_scale=0.2)

        exposure_time_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        background_sky_map = scaled_array.ScaledSquarePixelArray.single_value(
            value=1.0, pixel_scale=0.1, shape=image.shape
        )

        with pytest.raises(exc.DataException):
            simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
                image=image,
                psf=psf,
                pixel_scale=0.1,
                exposure_time=1.0,
                exposure_time_map=exposure_time_map,
                background_sky_map=background_sky_map,
                add_noise=True,
                noise_seed=1,
            )

    def test__from_deflections_and_source_galaxies__same_as_manual_calculation_using_tracer(
        self
    ):

        psf = ccd.PSF(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scale=1.0,
        )

        image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
            shape=(10, 10), pixel_scale=1.0, sub_grid_size=1
        )

        g0 = g.Galaxy(
            redshift=0.5, mass_profile=mp.SphericalIsothermal(einstein_radius=1.0)
        )

        g1 = g.Galaxy(redshift=1.0, light=lp.SphericalSersic(intensity=1.0))

        tracer = ray_tracing.TracerImageSourcePlanes(
            lens_galaxies=[g0],
            source_galaxies=[g1],
            image_plane_grid_stack=image_plane_grid_stack,
        )

        deflections = tracer.deflections(return_in_2d=True, return_binned=True)

        ccd_data_simulated_via_deflections = simulated_ccd.SimulatedCCDData.from_deflections_source_galaxies_and_exposure_arrays(
            deflections=deflections,
            pixel_scale=1.0,
            source_galaxies=[g1],
            exposure_time=10000.0,
            background_sky_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        tracer_profile_image_plane_image = tracer.profile_image_plane_image(
            return_in_2d=True, return_binned=True
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=tracer_profile_image_plane_image,
            pixel_scale=1.0,
            exposure_time=10000.0,
            background_sky_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        assert (
            ccd_data_simulated_via_deflections.image == ccd_data_simulated.image
        ).all()
        assert (ccd_data_simulated_via_deflections.psf == ccd_data_simulated.psf).all()
        assert (
            ccd_data_simulated_via_deflections.noise_map == ccd_data_simulated.noise_map
        ).all()
        assert (
            ccd_data_simulated_via_deflections.background_sky_map
            == ccd_data_simulated.background_sky_map
        ).all()
        assert (
            ccd_data_simulated_via_deflections.exposure_time_map
            == ccd_data_simulated.exposure_time_map
        ).all()

    def test__from_tracer__same_as_manual_tracer_input(self):

        psf = ccd.PSF(
            array=np.array([[0.0, 1.0, 0.0], [1.0, 2.0, 1.0], [0.0, 1.0, 0.0]]),
            pixel_scale=1.0,
        )

        image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
            shape=(20, 20), pixel_scale=0.05, sub_grid_size=1
        )

        lens_galaxy = g.Galaxy(
            redshift=0.5,
            light=lp.EllipticalSersic(intensity=1.0),
            mass=mp.EllipticalIsothermal(einstein_radius=1.6),
        )

        source_galaxy = g.Galaxy(redshift=1.0, light=lp.EllipticalSersic(intensity=0.3))

        tracer = ray_tracing.TracerImageSourcePlanes(
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
            image_plane_grid_stack=image_plane_grid_stack,
        )

        ccd_data_simulated_via_tracer = simulated_ccd.SimulatedCCDData.from_tracer_and_exposure_arrays(
            tracer=tracer,
            pixel_scale=0.1,
            exposure_time=10000.0,
            psf=psf,
            background_sky_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        ccd_data_simulated = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
            image=tracer.padded_profile_image_plane_image_2d_from_psf_shape(
                psf_shape=(3, 3)
            ),
            pixel_scale=0.1,
            exposure_time=10000.0,
            psf=psf,
            background_sky_level=100.0,
            add_noise=True,
            noise_seed=1,
        )

        assert (ccd_data_simulated_via_tracer.image == ccd_data_simulated.image).all()
        assert (ccd_data_simulated_via_tracer.psf == ccd_data_simulated.psf).all()
        assert (
            ccd_data_simulated_via_tracer.noise_map == ccd_data_simulated.noise_map
        ).all()
        assert (
            ccd_data_simulated_via_tracer.background_sky_map
            == ccd_data_simulated.background_sky_map
        ).all()
        assert (
            ccd_data_simulated_via_tracer.exposure_time_map
            == ccd_data_simulated.exposure_time_map
        ).all()


class TestSimulatePoissonNoise(object):
    def test__input_image_all_0s__exposure_time_all_1s__all_noise_values_are_0s(self):

        image = np.zeros((2, 2))

        exposure_time = scaled_array.ScaledSquarePixelArray.single_value(
            1.0, image.shape, pixel_scale=0.1
        )
        simulated_poisson_image = image + simulated_ccd.generate_poisson_noise(
            image, exposure_time, seed=1
        )

        assert simulated_poisson_image.shape == (2, 2)
        assert (simulated_poisson_image == np.zeros((2, 2))).all()

    def test__input_image_includes_10s__exposure_time_is_1s__gives_noise_values_near_1_to_5(
        self
    ):
        image = np.array([[10.0, 0.0], [0.0, 10.0]])

        exposure_time = scaled_array.ScaledSquarePixelArray.single_value(
            1.0, image.shape, pixel_scale=0.1
        )
        poisson_noise_map = simulated_ccd.generate_poisson_noise(
            image, exposure_time, seed=1
        )
        simulated_poisson_image = image + poisson_noise_map

        assert simulated_poisson_image.shape == (2, 2)

        # Use known noise_map_1d map for given seed.
        assert (
            poisson_noise_map == np.array([[(10.0 - 9.0), 0], [0, (10.0 - 6.0)]])
        ).all()
        assert (simulated_poisson_image == np.array([[11, 0], [0, 14]])).all()

        assert (simulated_poisson_image - poisson_noise_map == image).all()

    def test__input_image_is_all_10s__exposure_time_is_1s__gives_noise_values_near_1_to_5(
        self
    ):
        image = np.array([[10.0, 10.0], [10.0, 10.0]])

        exposure_time = scaled_array.ScaledSquarePixelArray.single_value(
            1.0, image.shape, pixel_scale=0.1
        )
        poisson_noise_map = simulated_ccd.generate_poisson_noise(
            image, exposure_time, seed=1
        )
        simulated_poisson_image = image + poisson_noise_map

        assert simulated_poisson_image.shape == (2, 2)

        # Use known noise_map_1d map for given seed.
        assert (poisson_noise_map == np.array([[1, 4], [3, 1]])).all()

        assert (simulated_poisson_image == np.array([[11, 14], [13, 11]])).all()

        assert (simulated_poisson_image - poisson_noise_map == image).all()

    def test__input_image_has_1000000s__exposure_times_is_1s__these_give_positive_noise_values_near_1000(
        self
    ):
        image = np.array([[10000000.0, 0.0], [0.0, 10000000.0]])

        exposure_time = scaled_array.ScaledSquarePixelArray(
            array=np.ones((2, 2)), pixel_scale=0.1
        )

        poisson_noise_map = simulated_ccd.generate_poisson_noise(
            image, exposure_time, seed=2
        )

        simulated_poisson_image = image + poisson_noise_map

        assert simulated_poisson_image.shape == (2, 2)

        # Use known noise_map_1d map for given seed.
        assert (poisson_noise_map == np.array([[571, 0], [0, -441]])).all()

        assert (
            simulated_poisson_image
            == np.array([[10000000.0 + 571, 0.0], [0.0, 10000000.0 - 441]])
        ).all()

        assert (simulated_poisson_image - poisson_noise_map == image).all()

    def test__two_images_same_in_counts_but_different_in_electrons_per_sec__noise_related_by_exposure_times(
        self
    ):
        image_0 = np.array([[10.0, 0.0], [0.0, 10.0]])

        exposure_time_0 = scaled_array.ScaledSquarePixelArray(
            array=np.ones((2, 2)), pixel_scale=0.1
        )

        image_1 = np.array([[5.0, 0.0], [0.0, 5.0]])

        exposure_time_1 = scaled_array.ScaledSquarePixelArray(
            array=2.0 * np.ones((2, 2)), pixel_scale=0.1
        )

        simulated_poisson_image_0 = image_0 + simulated_ccd.generate_poisson_noise(
            image_0, exposure_time_0, seed=1
        )
        simulated_poisson_image_1 = image_1 + simulated_ccd.generate_poisson_noise(
            image_1, exposure_time_1, seed=1
        )

        assert (simulated_poisson_image_0 / 2.0 == simulated_poisson_image_1).all()

    def test__same_as_above_but_range_of_image_values_and_exposure_times(self):
        image_0 = np.array([[10.0, 20.0], [30.0, 40.0]])

        exposure_time_0 = scaled_array.ScaledSquarePixelArray(
            array=np.array([[2.0, 2.0], [3.0, 4.0]]), pixel_scale=0.1
        )

        image_1 = np.array([[20.0, 20.0], [45.0, 20.0]])

        exposure_time_1 = scaled_array.ScaledSquarePixelArray(
            array=np.array([[1.0, 2.0], [2.0, 8.0]]), pixel_scale=0.1
        )

        simulated_poisson_image_0 = image_0 + simulated_ccd.generate_poisson_noise(
            image_0, exposure_time_0, seed=1
        )
        simulated_poisson_image_1 = image_1 + simulated_ccd.generate_poisson_noise(
            image_1, exposure_time_1, seed=1
        )

        assert (
            simulated_poisson_image_0[0, 0] == simulated_poisson_image_1[0, 0] / 2.0
        ).all()
        assert simulated_poisson_image_0[0, 1] == simulated_poisson_image_1[0, 1]
        assert (
            simulated_poisson_image_0[1, 0] * 1.5
            == pytest.approx(simulated_poisson_image_1[1, 0], 1e-2)
        ).all()
        assert (
            simulated_poisson_image_0[1, 1] / 2.0 == simulated_poisson_image_1[1, 1]
        ).all()



import autofit as af
from autolens.data import ccd
from autolens.data import simulated_ccd
from autolens.data.array import grids
from autolens.lens import ray_tracing
from autolens.model.galaxy import galaxy as g
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import ray_tracing_plotters

from test.simulation import simulation_util

import os


def simulate_image_from_galaxies_and_output_to_fits(
    data_resolution,
    data_type,
    sub_grid_size,
    lens_galaxies,
    source_galaxies,
    psf_shape=(51, 51),
    exposure_time=300.0,
    background_sky_level=1.0,
):

    pixel_scale = simulation_util.pixel_scale_from_data_resolution(
        data_resolution=data_resolution
    )
    shape = simulation_util.shape_from_data_resolution(data_resolution=data_resolution)

    # Simulate a simple Gaussian PSF for the image.
    psf = ccd.PSF.from_gaussian(
        shape=psf_shape, sigma=pixel_scale, pixel_scale=pixel_scale
    )

    # Setup the image-plane grid stack of the CCD array which will be used for generating the image-plane image of the
    # simulated strong lens. A high-res sub-grid is necessary to ensure we fully resolve the central regions of the
    # lens and source galaxy light.
    image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
        shape=shape, pixel_scale=pixel_scale, sub_grid_size=sub_grid_size
    )

    # Use the input galaxies to setup a tracer, which will generate the image-plane image for the simulated CCD data.
    tracer = ray_tracing.TracerImageSourcePlanes(
        lens_galaxies=lens_galaxies,
        source_galaxies=source_galaxies,
        image_plane_grid_stack=image_plane_grid_stack,
    )

    # Simulate the CCD data, remembering that we use a special image-plane image which ensures edge-effects don't
    # degrade our modeling of the telescope optics (e.g. the PSF convolution).
    simulated_ccd_data = simulated_ccd.SimulatedCCDData.from_image_and_exposure_arrays(
        image=tracer.padded_profile_image_plane_image_2d_from_psf_shape,
        pixel_scale=pixel_scale,
        psf=psf,
        exposure_time=exposure_time,
        background_sky_level=background_sky_level,
        add_noise=True,
    )

    # Now, lets output this simulated ccd-data to the test/data folder.
    test_path = "{}/../".format(os.path.dirname(os.path.realpath(__file__)))

    data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=test_path, folder_names=["data", data_type, data_resolution]
    )

    ccd.output_ccd_data_to_fits(
        ccd_data=simulated_ccd_data,
        image_path=data_path + "image.fits",
        psf_path=data_path + "psf.fits",
        noise_map_path=data_path + "noise_map.fits",
        overwrite=True,
    )

    ccd_plotters.plot_ccd_subplot(
        ccd_data=simulated_ccd_data,
        output_filename="ccd_data",
        output_path=data_path,
        output_format="png",
    )

    ccd_plotters.plot_ccd_individual(
        ccd_data=simulated_ccd_data,
        should_plot_image=True,
        should_plot_noise_map=True,
        should_plot_psf=True,
        should_plot_signal_to_noise_map=True,
        output_path=data_path,
        output_format="png",
    )

    ray_tracing_plotters.plot_ray_tracing_subplot(
        tracer=tracer,
        output_filename="tracer",
        output_path=data_path,
        output_format="png",
    )

    ray_tracing_plotters.plot_ray_tracing_individual(
        tracer=tracer,
        should_plot_image_plane_image=True,
        should_plot_source_plane=True,
        should_plot_convergence=True,
        should_plot_potential=True,
        should_plot_deflections=True,
        output_path=data_path,
        output_format="png",
    )


def make_lens_only_dev_vaucouleurs(data_resolutions, sub_grid_size):

    data_type = "lens_only_dev_vaucouleurs"

    # This lens-only system has a Dev Vaucouleurs spheroid / bulge.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        bulge=lp.EllipticalDevVaucouleurs(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.1,
            effective_radius=1.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[g.Galaxy(redshift=1.0)],
        )


def make_lens_only_bulge_and_disk(data_resolutions, sub_grid_size):

    data_type = "lens_only_bulge_and_disk"

    # This source-only system has a Dev Vaucouleurs spheroid / bulge and surrounding Exponential envelope

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        bulge=lp.EllipticalDevVaucouleurs(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.1,
            effective_radius=1.0,
        ),
        envelope=lp.EllipticalExponential(
            centre=(0.0, 0.0),
            axis_ratio=0.7,
            phi=60.0,
            intensity=1.0,
            effective_radius=2.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[g.Galaxy(redshift=1.0)],
        )


def make_lens_only_x2_galaxies(data_resolutions, sub_grid_size):

    data_type = "lens_only_x2_galaxies"

    # This source-only system has two Sersic bulges separated by 2.0"

    lens_galaxy_0 = g.Galaxy(
        redshift=0.5,
        bulge=lp.EllipticalSersic(
            centre=(-1.0, -1.0),
            axis_ratio=0.8,
            phi=0.0,
            intensity=1.0,
            effective_radius=1.0,
            sersic_index=3.0,
        ),
    )

    lens_galaxy_1 = g.Galaxy(
        redshift=0.5,
        bulge=lp.EllipticalSersic(
            centre=(1.0, 1.0),
            axis_ratio=0.8,
            phi=0.0,
            intensity=1.0,
            effective_radius=1.0,
            sersic_index=3.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy_0, lens_galaxy_1],
            source_galaxies=[g.Galaxy(redshift=1.0)],
        )


def make_no_lens_light_and_source_smooth(data_resolutions, sub_grid_size):

    data_type = "no_lens_light_and_source_smooth"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        mass=mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )

    source_galaxy = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=60.0,
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
        )


def make_no_lens_light_and_source_cuspy(data_resolutions, sub_grid_size):

    data_type = "no_lens_light_and_source_cuspy"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        mass=mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )

    source_galaxy = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=60.0,
            intensity=0.1,
            effective_radius=0.5,
            sersic_index=3.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
        )


def make_no_lens_light_and_source_smooth_offset_centre(data_resolutions, sub_grid_size):

    data_type = "no_lens_light_and_source_smooth_offset_centre"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        mass=mp.EllipticalIsothermal(
            centre=(1.0, 1.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )

    source_galaxy = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(1.0, 1.0),
            axis_ratio=0.8,
            phi=60.0,
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
        )


def make_lens_light_and_source_smooth(data_resolutions, sub_grid_size):

    data_type = "lens_light_and_source_smooth"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.5,
            effective_radius=0.8,
            sersic_index=4.0,
        ),
        mass=mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )

    source_galaxy = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=60.0,
            intensity=0.4,
            effective_radius=0.5,
            sersic_index=1.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
        )


def make_lens_light_and_source_cuspy(data_resolutions, sub_grid_size):

    data_type = "lens_light_and_source_cuspy"

    # This source-only system has a smooth source (low Sersic Index) and simple SIE mass profile.

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.9,
            phi=45.0,
            intensity=0.5,
            effective_radius=0.8,
            sersic_index=4.0,
        ),
        mass=mp.EllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=1.6, axis_ratio=0.7, phi=45.0
        ),
    )

    source_galaxy = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.0, 0.0),
            axis_ratio=0.8,
            phi=60.0,
            intensity=0.1,
            effective_radius=0.5,
            sersic_index=3.0,
        ),
    )

    for data_resolution in data_resolutions:

        simulate_image_from_galaxies_and_output_to_fits(
            data_type=data_type,
            data_resolution=data_resolution,
            sub_grid_size=sub_grid_size,
            lens_galaxies=[lens_galaxy],
            source_galaxies=[source_galaxy],
        )



import os

import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from test.integration import integration_util
from test.simulation import simulation_util

test_type = "grid_search"
test_name = "multinest_grid_subhalo"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(
        data_type="no_lens_light_and_source_smooth", data_resolution="LSST"
    )
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data)


def make_pipeline(test_name):
    class QuickPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.mass.centre_1 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.lens_galaxies.lens.mass.axis_ratio = af.UniformPrior(
                lower_limit=0.65, upper_limit=0.75
            )
            self.lens_galaxies.lens.mass.phi = af.UniformPrior(
                lower_limit=40.0, upper_limit=50.0
            )
            self.lens_galaxies.lens.mass.einstein_radius = af.UniformPrior(
                lower_limit=1.55, upper_limit=1.65
            )

            self.source_galaxies.source.light.centre_0 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.source_galaxies.source.light.centre_1 = af.UniformPrior(
                lower_limit=-0.01, upper_limit=0.01
            )
            self.source_galaxies.source.light.axis_ratio = af.UniformPrior(
                lower_limit=0.75, upper_limit=0.85
            )
            self.source_galaxies.source.light.phi = af.UniformPrior(
                lower_limit=50.0, upper_limit=70.0
            )
            self.source_galaxies.source.light.intensity = af.UniformPrior(
                lower_limit=0.35, upper_limit=0.45
            )
            self.source_galaxies.source.light.effective_radius = af.UniformPrior(
                lower_limit=0.45, upper_limit=0.55
            )
            self.source_galaxies.source.light.sersic_index = af.UniformPrior(
                lower_limit=0.9, upper_limit=1.1
            )

    phase1 = QuickPhase(
        phase_name="phase_1",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal)
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    class GridPhase(af.phase.as_grid_search(phase_imaging.LensSourcePlanePhase)):
        @property
        def grid_priors(self):
            return [
                self.variable.lens_galaxies.subhalo.mass.centre_0,
                self.variable.lens_galaxies.subhalo.mass.centre_1,
            ]

        def pass_priors(self, results):

            ### Lens Mass, PL -> PL, Shear -> Shear ###

            # self.lens_galaxies.lens = results.from_phase('phase_1').\
            #     constant.lens_galaxies.lens

            ### Lens Subhalo, Adjust priors to physical masses (10^6 - 10^10) and concentrations (6-24)

            self.lens_galaxies.subhalo.mass.kappa_s = af.UniformPrior(
                lower_limit=0.0005, upper_limit=0.2
            )
            self.lens_galaxies.subhalo.mass.scale_radius = af.UniformPrior(
                lower_limit=0.001, upper_limit=1.0
            )
            self.lens_galaxies.subhalo.mass.centre_0 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )
            self.lens_galaxies.subhalo.mass.centre_1 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )

            ### Source Light, Sersic -> Sersic ###

            self.source_galaxies.source.light.centre = (
                results.from_phase("phase_1")
                .variable_absolute(a=0.05)
                .source_galaxies.source.light.centre
            )

            self.source_galaxies.source.light.intensity = (
                results.from_phase("phase_1")
                .variable_relative(r=0.5)
                .source_galaxies.source.light.intensity
            )

            self.source_galaxies.source.light.effective_radius = (
                results.from_phase("phase_1")
                .variable_relative(r=0.5)
                .source_galaxies.source.light.effective_radius
            )

            self.source_galaxies.source.light.sersic_index = (
                results.from_phase("phase_1")
                .variable_relative(r=0.5)
                .source_galaxies.source.light.sersic_index
            )

            self.source_galaxies.source.light.axis_ratio = (
                results.from_phase("phase_1")
                .variable_absolute(a=0.1)
                .source_galaxies.source.light.axis_ratio
            )

            self.source_galaxies.source.light.phi = (
                results.from_phase("phase_1")
                .variable_absolute(a=20.0)
                .source_galaxies.source.light.phi
            )

    phase2 = GridPhase(
        phase_name="phase_2",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, mass=mp.EllipticalIsothermal),
            subhalo=gm.GalaxyModel(
                redshift=0.5, mass=mp.SphericalTruncatedNFWChallenge
            ),
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        optimizer_class=af.MultiNest,
        number_of_steps=2,
    )

    phase2.optimizer.const_efficiency_mode = True

    return pl.PipelineImaging(test_name, phase1, phase2)


if __name__ == "__main__":
    pipeline()



import autofit as af
from autolens.data.array import mask as msk
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp


# In this pipeline, we'll demonstrate binning up - which allows us to fit a binned up version of an image in a phse of
# the pipeline. In this example, we will perform an initial analysis on an image binned up from a pixel scale of 0.05"
# to 0.1", which gives significant speed-up in run timme, and then refine the model in a second phase at the input
# resolution.

# Whilst bin up factors can be manually specified in the pipeline, in this example we will make the bin up factor
# an input parameter of the pipeline. This means we can run the pipeline with different binning up factors for different
# runners.

# We will also use phase tagging to ensure phases which use binned up data have a tag in their path, so it is clear
# what settings a phases has when it uses this feature.

# We'll perform a basic analysis which fits a lensed source galaxy using a parametric light profile where
# the lens's light is omitted. This pipeline uses two phases:

# Phase 1:

# Description: initialize the lens mass model and source light profile using x1 source with a bin up factor of x2.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses a bin up factor of x2

# Phase 2:

# Description: Fits the lens and source model using unbinned data.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens mass (variable -> phase 1), source light (variable -> phase 1)
# Notes: No binning up.


def make_pipeline(phase_folders=None, bin_up_factor=2):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pl__binning_up"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name
    )

    # When a phase is passed a bin_up_factor, a settings tag is automatically generated and added to the phase path,
    # to make it clear what binning up was used. The settings tag, phase name and phase paths are shown for 3
    # example bin up factors:

    # bin_up_factor=2 -> phase_path=phase_name/settings_bin_up_2
    # bin_up_factor=3 -> phase_path=phase_name/settings_bin_up_3

    # If the bin_up_facor is None or 1, the tag is an empty string, thus not changing the settings tag:

    # bin_up_factor=None -> phase_path=phase_name/settings
    # bin_up_factor=1 -> phase_path=phase_name/settings

    # This function uses the phase folders and pipeline name to set up the output directory structure,
    # e.g. 'autolens_workspace/output/phase_folder_1/phase_folder_2/pipeline_name/phase_name/settings_tag/'

    phase_folders.append(pipeline_name)

    # As there is no lens light component, we can use an annular mask throughout this pipeline which removes the
    # central regions of the image.

    def mask_function(image):
        return msk.Mask.circular_annular(
            shape=image.shape,
            pixel_scale=image.pixel_scale,
            inner_radius_arcsec=0.2,
            outer_radius_arcsec=3.3,
        )

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Bin up the image by the input factor specified, which is default 2.

    class LensSourceX1Phase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens.mass.centre_0 = af.GaussianPrior(
                mean=0.0, sigma=0.1
            )
            self.lens_galaxies.lens.mass.centre_1 = af.GaussianPrior(
                mean=0.0, sigma=0.1
            )

    phase1 = LensSourceX1Phase(
        phase_name="phase_1_x1_source",
        phase_folders=phase_folders,
        tag_phases=True,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, mass=mp.EllipticalIsothermal, shear=mp.ExternalShear
            )
        ),
        source_galaxies=dict(
            source_0=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        mask_function=mask_function,
        bin_up_factor=bin_up_factor,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and two source galaxies, where we:

    # 1) Omit the bin up factor, thus performing the modeling at the image's native resolution.

    class LensSourceX2Phase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            self.lens_galaxies.lens = results.from_phase(
                "phase_1_x1_source"
            ).variable.lens_galaxies.lens

            self.source_galaxies.source = results.from_phase(
                "phase_1_x1_source"
            ).variable.source_galaxies.source

    phase2 = LensSourceX2Phase(
        phase_name="phase_2_x2_source",
        phase_folders=phase_folders,
        tag_phases=True,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5, mass=mp.EllipticalIsothermal, shear=mp.ExternalShear
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        mask_function=mask_function,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2)



from autolens.pipeline import tagging

from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg


class TestPipelineNameTag:
    def test__pipeline_name_and_tag__mixture_of_values(self):

        pipeline_name = tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl", fix_lens_light=True
        )

        assert pipeline_name == "pl_fix_lens_light"

        pipeline_tag = tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl",
            fix_lens_light=True,
            pixelization=pix.Rectangular,
            regularization=reg.Constant,
        )

        assert pipeline_tag == "pl_fix_lens_light_pix_rect_reg_const"

        pipeline_name = tagging.pipeline_name_from_name_and_settings(
            pipeline_name="pl2", fix_lens_light=True, align_bulge_disk_phi=True
        )

        assert pipeline_name == "pl2_fix_lens_light_bd_align_phi"

    def test__pipeline_tag__mixture_of_values(self):

        pipeline_tag = tagging.pipeline_tag_from_pipeline_settings(fix_lens_light=True)

        assert pipeline_tag == "_fix_lens_light"

        pipeline_tag = tagging.pipeline_tag_from_pipeline_settings(
            fix_lens_light=True,
            pixelization=pix.Rectangular,
            regularization=reg.Constant,
        )

        assert pipeline_tag == "_fix_lens_light_pix_rect_reg_const"

        pipeline_tag = tagging.pipeline_tag_from_pipeline_settings(
            fix_lens_light=True, align_bulge_disk_phi=True
        )

        assert pipeline_tag == "_fix_lens_light_bd_align_phi"


class TestPipelineTaggers:
    def test__fix_lens_light_tagger(self):
        tag = tagging.fix_lens_light_tag_from_fix_lens_light(fix_lens_light=False)
        assert tag == ""
        tag = tagging.fix_lens_light_tag_from_fix_lens_light(fix_lens_light=True)
        assert tag == "_fix_lens_light"

    def test__pixelization_tagger(self):

        tag = tagging.pixelization_tag_from_pixelization(pixelization=None)
        assert tag == ""
        tag = tagging.pixelization_tag_from_pixelization(pixelization=pix.Rectangular)
        assert tag == "_pix_rect"
        tag = tagging.pixelization_tag_from_pixelization(
            pixelization=pix.VoronoiBrightnessImage
        )
        assert tag == "_pix_voro_image"

    def test__regularization_tagger(self):

        tag = tagging.regularization_tag_from_regularization(regularization=None)
        assert tag == ""
        tag = tagging.regularization_tag_from_regularization(
            regularization=reg.Constant
        )
        assert tag == "_reg_const"
        tag = tagging.regularization_tag_from_regularization(
            regularization=reg.AdaptiveBrightness
        )
        assert tag == "_reg_adapt_bright"

    def test__align_bulge_disk_taggers(self):
        tag = tagging.align_bulge_disk_centre_tag_from_align_bulge_disk_centre(
            align_bulge_disk_centre=False
        )
        assert tag == ""
        tag = tagging.align_bulge_disk_centre_tag_from_align_bulge_disk_centre(
            align_bulge_disk_centre=True
        )
        assert tag == "_bd_align_centre"

        tag = tagging.align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
            align_bulge_disk_axis_ratio=False
        )
        assert tag == ""
        tag = tagging.align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
            align_bulge_disk_axis_ratio=True
        )
        assert tag == "_bd_align_axis_ratio"

        tag = tagging.align_bulge_disk_phi_tag_from_align_bulge_disk_phi(
            align_bulge_disk_phi=False
        )
        assert tag == ""
        tag = tagging.align_bulge_disk_phi_tag_from_align_bulge_disk_phi(
            align_bulge_disk_phi=True
        )
        assert tag == "_bd_align_phi"

    def test__bulge_disk_tag(self):
        tag = tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=False,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert tag == ""

        tag = tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=False,
        )
        assert tag == "_bd_align_centre"

        tag = tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=False,
            align_bulge_disk_phi=True,
        )
        assert tag == "_bd_align_centre_bd_align_phi"

        tag = tagging.bulge_disk_tag_from_align_bulge_disks(
            align_bulge_disk_centre=True,
            align_bulge_disk_axis_ratio=True,
            align_bulge_disk_phi=True,
        )
        assert tag == "_bd_align_centre_bd_align_axis_ratio_bd_align_phi"


class TestPhaseTag:
    def test__mixture_of_values(self):

        phase_tag = tagging.phase_tag_from_phase_settings(
            sub_grid_size=2,
            bin_up_factor=None,
            image_psf_shape=None,
            inversion_psf_shape=None,
            inner_mask_radii=0.3,
            positions_threshold=2.0,
            interp_pixel_scale=None,
            cluster_pixel_scale=None,
        )

        assert phase_tag == "_sub_2_pos_2.00_inner_mask_0.30"

        phase_tag = tagging.phase_tag_from_phase_settings(
            sub_grid_size=1,
            bin_up_factor=3,
            image_psf_shape=(2, 2),
            inversion_psf_shape=(3, 3),
            inner_mask_radii=None,
            positions_threshold=None,
            interp_pixel_scale=0.2,
            cluster_pixel_scale=0.3,
        )

        assert (
            phase_tag
            == "_sub_1_bin_up_3_image_psf_2x2_inv_psf_3x3_interp_0.200_cluster_0.300"
        )


class TestPhaseTaggers:
    def test__positions_threshold_tagger(self):

        tag = tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=None
        )
        assert tag == ""
        tag = tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=1.0
        )
        assert tag == "_pos_1.00"
        tag = tagging.positions_threshold_tag_from_positions_threshold(
            positions_threshold=2.56
        )
        assert tag == "_pos_2.56"

    def test__inner_circular_mask_radii_tagger(self):

        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=None
        )
        assert tag == ""
        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=0.2
        )
        print(tag)
        assert tag == "_inner_mask_0.20"
        tag = tagging.inner_mask_radii_tag_from_inner_circular_mask_radii(
            inner_mask_radii=3
        )
        assert tag == "_inner_mask_3.00"

    def test__sub_grid_size_tagger(self):

        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=1)
        assert tag == "_sub_1"
        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=2)
        assert tag == "_sub_2"
        tag = tagging.sub_grid_size_tag_from_sub_grid_size(sub_grid_size=4)
        assert tag == "_sub_4"

    def test__bin_up_factor_tagger(self):

        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=None)
        assert tag == ""
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=1)
        assert tag == ""
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=2)
        assert tag == "_bin_up_2"
        tag = tagging.bin_up_factor_tag_from_bin_up_factor(bin_up_factor=3)
        assert tag == "_bin_up_3"

    def test__image_psf_shape_tagger(self):

        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=None)
        assert tag == ""
        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=(2, 2))
        assert tag == "_image_psf_2x2"
        tag = tagging.image_psf_shape_tag_from_image_psf_shape(image_psf_shape=(3, 4))
        assert tag == "_image_psf_3x4"

    def test__inversion_psf_shape_tagger(self):

        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=None
        )
        assert tag == ""
        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(2, 2)
        )
        assert tag == "_inv_psf_2x2"
        tag = tagging.inversion_psf_shape_tag_from_inversion_psf_shape(
            inversion_psf_shape=(3, 4)
        )
        assert tag == "_inv_psf_3x4"

    def test__interp_pixel_scale_tagger(self):

        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=None
        )
        assert tag == ""
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.5
        )
        assert tag == "_interp_0.500"
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.25
        )
        assert tag == "_interp_0.250"
        tag = tagging.interp_pixel_scale_tag_from_interp_pixel_scale(
            interp_pixel_scale=0.234
        )
        assert tag == "_interp_0.234"

    def test__cluster_pixel_scale_tagger(self):

        tag = tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=None
        )
        assert tag == ""
        tag = tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.5
        )
        assert tag == "_cluster_0.500"
        tag = tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.25
        )
        assert tag == "_cluster_0.250"
        tag = tagging.cluster_pixel_scale_tag_from_cluster_pixel_scale(
            cluster_pixel_scale=0.234
        )
        assert tag == "_cluster_0.234"



import os

import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from autolens.model.hyper import hyper_data as hi
from autolens.model.profiles import light_profiles as lp
from test.integration import integration_util
from test.simulation import simulation_util

test_type = "lens_only"
test_name = "lens_x1_galaxy_hyper_image_noise"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)


def pipeline():

    integration_util.reset_paths(test_name=test_name, output_path=output_path)

    ccd_data = simulation_util.load_test_ccd_data(
        data_type="lens_only_dev_vaucouleurs", data_resolution="LSST"
    )

    pipeline = make_pipeline(test_name=test_name)

    pipeline.run(data=ccd_data)


def make_pipeline(test_name):

    phase1 = phase_imaging.LensPlanePhase(
        phase_name="phase_1",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(redshift=0.5, sersic=lp.EllipticalSersic)
        ),
        hyper_image_sky=hi.HyperImageSky,
        hyper_noise_background=hi.HyperNoiseBackground,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(test_name, phase1)


if __name__ == "__main__":
    pipeline()



import autofit as af
from autolens.data import ccd
from autolens.data import simulated_ccd
from autolens.data.array import mask as msk
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.model.galaxy import galaxy as g
from autolens.lens import lens_data as ld
from autolens.lens import ray_tracing
from autolens.lens import lens_fit
from autolens.data.plotters import ccd_plotters
from autolens.lens.plotters import lens_fit_plotters

import os

# So far, we've not paid much attention to the source galaxy's morphology. We've assumed its a single-component
# exponential profile, which is a fairly crude assumption. A quick look at any image of a real galaxy reveals a wealth
# of different structures that could be present - bulges, disks, bars, star-forming knots and so on.
# Furthermore, there could be more than one source-galaxy!

# In this example, we'll explore how far we can get trying to_fit a complex source using a pipeline. Fitting complex
# source's is an exercise in diminishing returns. Each component we add to our source model brings with it an extra 5-7,
# parameters. If there are 4 components, or multiple galaxies, we're quickly entering the somewhat nasty regime of
# 30-40+ parameters in our non-linear search. Even with a pipeline, thats a lot of parameters to try and fit!

# Lets setup the path to the workspace, as per usual.
workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

# This function simulates an image with a complex source.
def simulate():

    from autolens.data.array import grids
    from autolens.model.galaxy import galaxy as g
    from autolens.lens import ray_tracing

    psf = ccd.PSF.from_gaussian(shape=(11, 11), sigma=0.05, pixel_scale=0.05)

    image_plane_grid_stack = grids.GridStack.from_shape_pixel_scale_and_sub_grid_size(
        shape=(180, 180), pixel_scale=0.05, psf_shape=(11, 11)
    )

    lens_galaxy = g.Galaxy(
        redshift=0.5,
        mass=mp.EllipticalIsothermal(
            centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
        ),
    )

    source_galaxy_0 = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.1, 0.1),
            axis_ratio=0.8,
            phi=90.0,
            intensity=0.2,
            effective_radius=1.0,
            sersic_index=1.5,
        ),
    )

    source_galaxy_1 = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(-0.25, 0.25),
            axis_ratio=0.7,
            phi=45.0,
            intensity=0.1,
            effective_radius=0.2,
            sersic_index=3.0,
        ),
    )

    source_galaxy_2 = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(0.45, -0.35),
            axis_ratio=0.6,
            phi=90.0,
            intensity=0.03,
            effective_radius=0.3,
            sersic_index=3.5,
        ),
    )

    source_galaxy_3 = g.Galaxy(
        redshift=1.0,
        light=lp.EllipticalSersic(
            centre=(-0.05, -0.0),
            axis_ratio=0.9,
            phi=140.0,
            intensity=0.03,
            effective_radius=0.1,
            sersic_index=4.0,
        ),
    )

    tracer = ray_tracing.TracerImageSourcePlanes(
        lens_galaxies=[lens_galaxy],
        source_galaxies=[
            source_galaxy_0,
            source_galaxy_1,
            source_galaxy_2,
            source_galaxy_3,
        ],
        image_plane_grid_stack=image_plane_grid_stack,
    )

    return simulated_ccd.SimulatedCCDData.from_tracer_and_exposure_arrays(
        tracer=tracer,
        pixel_scale=0.05,
        exposure_time=300.0,
        psf=psf,
        background_sky_level=0.1,
        add_noise=True,
    )


# Lets simulate the image.
ccd_data = simulate()
ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# Yep, that's a pretty complex source. There are clearly more than 4 peaks of light - I wouldn't like to guess whether
# there is one or two sources (or more). You'll also notice I omitted the lens galaxy's light for this system, this is
# to keep the number of parameters down and the phases running fast, but we wouldn't get such a luxury in a real galaxy.

# Again, before we checkout the pipeline, lets import it, and get it running.
from workspace.howtolens.chapter_3_pipelines import tutorial_3_pipeline_complex_source

pipeline_complex_source = tutorial_3_pipeline_complex_source.make_pipeline(
    phase_folders=["howtolens", "c3_t3_complex_source"]
)

pipeline_complex_source.run(data=ccd_data)

# Okay, so with 4 sources, we still couldn't get a good a fit to the source that didn't leave residuals. The thing is,
# I did infact simulate the lens with 4 sources. This means that there is a 'perfect fit', somewhere in that parameter
# space, that we unfortunately missed using the pipeline above.
#
# Lets confirm this, by manually fitting the ccd data with the true input model.

lens_data = ld.LensData(
    ccd_data=ccd_data,
    mask=msk.Mask.circular(
        shape=ccd_data.shape, pixel_scale=ccd_data.pixel_scale, radius_arcsec=3.0
    ),
)

lens_galaxy = g.Galaxy(
    redshift=0.5,
    mass=mp.EllipticalIsothermal(
        centre=(0.0, 0.0), axis_ratio=0.8, phi=135.0, einstein_radius=1.6
    ),
)

source_galaxy_0 = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.1, 0.1),
        axis_ratio=0.8,
        phi=90.0,
        intensity=0.2,
        effective_radius=1.0,
        sersic_index=1.5,
    ),
)

source_galaxy_1 = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(-0.25, 0.25),
        axis_ratio=0.7,
        phi=45.0,
        intensity=0.1,
        effective_radius=0.2,
        sersic_index=3.0,
    ),
)

source_galaxy_2 = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(0.45, -0.35),
        axis_ratio=0.6,
        phi=90.0,
        intensity=0.03,
        effective_radius=0.3,
        sersic_index=3.5,
    ),
)

source_galaxy_3 = g.Galaxy(
    redshift=1.0,
    light=lp.EllipticalSersic(
        centre=(-0.05, -0.0),
        axis_ratio=0.9,
        phi=140.0,
        intensity=0.03,
        effective_radius=0.1,
        sersic_index=4.0,
    ),
)

tracer = ray_tracing.TracerImageSourcePlanes(
    lens_galaxies=[lens_galaxy],
    source_galaxies=[
        source_galaxy_0,
        source_galaxy_1,
        source_galaxy_2,
        source_galaxy_3,
    ],
    image_plane_grid_stack=lens_data.grid_stack,
)

true_fit = lens_fit.LensDataFit.for_data_and_tracer(lens_data=lens_data, tracer=tracer)

lens_fit_plotters.plot_fit_subplot(
    fit=true_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

lens_fit_plotters.plot_fit_subplot_of_planes(
    fit=true_fit,
    should_plot_mask=True,
    extract_array_from_mask=True,
    zoom_around_mask=True,
)

# And indeed, we see far improved residuals, chi-squareds, etc.

# The morale of this story is that, if the source morphology is complex, there is no way we can build a pipeline to
# fit it. For this tutorial, this was true even though our source model could actually fit the data perfectly. For
# real lenses, the source will be *even more complex* and there is even less hope of getting a good fit :(

# But fear not, PyAutoLens has you covered. In chapter 4, we'll introduce a completely new way to model the source
# galaxy, which addresses the problem faced here. But before that, in the next tutorial we'll discuss how we actually
# pass priors in a pipeline.



from autofit import conf

from autolens.data import ccd
from autolens.data.array import mask as msk

import os

# Given your username is where your data is stored, you'll need to put your cosma username here.
cosma_username = "pdtw24"
cosma_path = "/cosma7/data/dp004/dc-nigh1/autolens/"

data_folder = "subhalo_challenge"

cosma_data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["data", data_folder]
)
cosma_output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    cosma_path, folder_names=["output", data_folder]
)

workspace_path = "{}/../../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path, and override the output path with the Cosma path.
conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=cosma_output_path
)

# The fifth line of this batch script - '#SBATCH --array=1-17' is what species this. Its telling Cosma we're going to
# run 17 jobs, and the id's of those jobs will be numbered from 1 to 17. Infact, these ids are passed to this runner,
# and we'll use them to ensure that each jobs loads a different image. Lets get the cosma array id for our job.
# cosma_array_id = int(sys.argv[1])
# data_level = str(sys.argv[2])
# data_name = str(sys.argv[3])

data_type = "noise_normal"
data_level = "level_0"
data_name = "large_hi_sn_system_1"

pixel_scale = 0.00976562

resized_ccd_shape = (550, 550)

data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=cosma_data_path, folder_names=[data_type, data_level, data_name]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
    resized_ccd_shape=resized_ccd_shape,
)

mask = msk.load_mask_from_fits(
    mask_path=data_path + "mask_irregular.fits", pixel_scale=pixel_scale
)
mask = mask.resized_scaled_array_from_array(new_shape=resized_ccd_shape)

# from workspace_jam.pipelines.no_lens_light.initialize import lens_sie_shear_source_sersic
# from workspace_jam.pipelines.no_lens_light.power_law.from_initialize import lens_pl_shear_source_sersic
# from workspace_jam.pipelines.no_lens_light.subhalo.from_power_law import lens_pl_shear_subhalo_source_sersic
#
# pipeline_initialize = lens_sie_shear_source_sersic.make_pipeline(phase_folders=[data_type, data_level, data_name])
# pipeline_power_law = lens_pl_shear_source_sersic.make_pipeline(phase_folders=[data_type, data_level, data_name])
# pipeline_subhalo = lens_pl_shear_subhalo_source_sersic.make_pipeline(phase_folders=[data_type, data_level, data_name],
#                                                                      parallel=True)
#
# pipeline = pipeline_initialize + pipeline_power_law + pipeline_subhalo

from workspace_jam.pipelines.advanced.no_lens_light.initialize import (
    lens_sie_shear_source_sersic,
)
from workspace_jam.pipelines.advanced.no_lens_light.subhalo.from_initialize import (
    lens_sie_shear_subhalo_source_sersic,
)

pipeline_initialize = lens_sie_shear_source_sersic.make_pipeline(
    phase_folders=[data_type, data_level, data_name]
)
pipeline_subhalo = lens_sie_shear_subhalo_source_sersic.make_pipeline(
    phase_folders=[data_type, data_level, data_name], parallel=True
)

pipeline = pipeline_initialize + pipeline_subhalo

pipeline.run(data=ccd_data, mask=mask)



import autofit as af
from autolens.data import ccd
from autolens.data import mask as msk
from autolens.data.plotters import ccd_plotters

import os

# Welcome to the pipeline runner. This tool allows you to load data on strong lenses, and pass it to pipelines for a
# strong lens analysis. To show you around, we'll load up some example data and run it through some of the example
# pipelines that come distributed with PyAutoLens, in the 'autolens_workspace/pipelines/simple' folder.

# In this runner, our analysis assumes there is a lens light component that requires modeling and subtraction.

# Why is this folder called 'simple'? Well, we actually recommend you break pipelines up. Details of how to do this
# can be found in the 'autolens_workspace/runners/advanced' folder, but its a bit conceptually difficult. So, to
# familiarize yourself with PyAutoLens, I'd stick to the simple pipelines for now!

# You should also checkout the 'autolens_workspace/runners/features' folder. Here, you'll find features that are
# available in pipelines, that allow you to customize the analysis.

# This runner is supplied as both this Python script and a Juypter notebook. Its up to you which you use - I personally
# prefer the python script as provided you keep it relatively small, its quick and easy to comment out different lens
# names and pipelines to set off different analyses. However, notebooks are a tidier way to manage visualization - so
# feel free to use notebooks. Or, use both for a bit, and decide your favourite!

# Setup the path to the workspace, using a relative directory name.
workspace_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))

# Use this path to explicitly set the config path and output path.
af.conf.instance = af.conf.Config(
    config_path=workspace_path + "config", output_path=workspace_path + "output"
)

# It is convenient to specify the lens name as a string, so that if the pipeline is applied to multiple images we \
# don't have to change all of the path entries in the load_ccd_data_from_fits function below.

data_type = "example"
data_name = (
    "lens_light_mass_and_x1_source"
)  # An example simulated image with lens light emission and a source galaxy.
pixel_scale = 0.1

# Create the path where the data will be loaded from, which in this case is
# '/workspace/data/example/lens_light_mass_and_x1_source/'
data_path = af.path_util.make_and_return_path_from_path_and_folder_names(
    path=workspace_path, folder_names=["data", data_type, data_name]
)

ccd_data = ccd.load_ccd_data_from_fits(
    image_path=data_path + "image.fits",
    psf_path=data_path + "psf.fits",
    noise_map_path=data_path + "noise_map.fits",
    pixel_scale=pixel_scale,
)

ccd_plotters.plot_ccd_subplot(ccd_data=ccd_data)

# The example autolens_workspace data comes with a mask already, if you look in
# workspace/data/example/lens_light_and_x1_source/ you'll see a mask.fits file!
mask = msk.load_mask_from_fits(
    mask_path=data_path + "mask.fits", pixel_scale=pixel_scale
)

# Running a pipeline is easy, we simply import it from the pipelines folder and pass the lens data to its run function.
# Below, we'll use a 3 phase example pipeline to fit the data with a parametric lens light, mass and source light
# profile. Checkout _workspace/pipelines/examples/lens_sersic_sie_shear_source_sersic.py_' for a full description of
# the pipeline.

from workspace.pipelines.simple import lens_sersic_sie_shear_source_sersic

pipeline = lens_sersic_sie_shear_source_sersic.make_pipeline(
    phase_folders=[data_type, data_name]
)

pipeline.run(data=ccd_data)

# Another example pipeline is shown below, which fits the data using a pixelized inversion for the source light.
# If you commented out the 3 lines of code above, and uncomment the code below, you can easily set this pipeline
# off running!

# from workspace.pipelines.simple import lens_sersic_sie_shear_source_inversion
# pipeline = lens_sersic_sie_shear_source_inversion.make_pipeline(phase_folders=[data_type, data_name])
# pipeline.run(data=ccd_data)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import mass_profiles as mp
from autolens.model.inversion import pixelizations as pix
from autolens.model.inversion import regularization as reg

import os

# In this pipeline, we'll perform an analysis which fits an image with no lens light, and a source galaxy using an
# inversion, using a power-law mass profile. The pipeline follows on from the inversion pipeline
# ''pipelines/no_lens_light/inversion/lens_sie_shear_source_inversion_from_initialize.py'.

# The pipeline is two phases, as follows:

# Phase 1:

# Description: Fits the lens mass model as a power-law, using an inversion for the Source.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: no_lens_light/inversion/lens_sie_shear_source_inversion_from_initialize.py
# Prior Passing: Lens Mass (variable -> previous pipeline), source inversion (variable / constant -> previous pipeline)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations.

# Phase 2:

# Description: Refines the inversion parameters, using a fixed mass model from phase 1.
# Lens Mass: EllipitcalPowerLaw + ExternalShear
# Source Light: VoronoiMagnification + Constant
# Previous Pipelines: None
# Prior Passing: Lens Mass (constant -> phase 1), source inversion (variable -> phase 1)
# Notes: Uses an interpolation pixel scale for fast power-law deflection angle calculations.


def make_pipeline(
    pl_pixelization=pix.VoronoiMagnification,
    pl_regularization=reg.AdaptiveBrightness,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=0.05,
    inversion_pixel_limit=None,
    cluster_pixel_scale=0.1,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_pl__lens_pl_shear_source_inversion"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name,
        pixelization=pl_pixelization,
        regularization=pl_regularization,
    )

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy mass using the EllipticalIsothermal fit of the previous pipeline, and
    #    source inversion of the previous pipeline.

    class LensSourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ### Lens Mass, SIE -> PL ###

            self.lens_galaxies.lens.mass.centre = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens_galaxies.lens.mass.centre

            self.lens_galaxies.lens.mass.axis_ratio = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens_galaxies.lens.mass.axis_ratio

            self.lens_galaxies.lens.mass.phi = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens_galaxies.lens.mass.phi

            self.lens_galaxies.lens.mass.einstein_radius = (
                results.from_phase("phase_2_lens_sie_shear_source_inversion")
                .variable_absolute(a=0.3)
                .lens_galaxies.lens.mass.einstein_radius
            )

            ### Lens Shear, Shear -> Shear ###

            self.lens_galaxies.lens.phase_2_lens_sie_shear_source_magnification_inversion = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).variable.lens_galaxies.lens.shear

            ### Source Inversion, Inv -> Inv ###

            self.source_galaxies.source = results.from_phase(
                "phase_2_lens_sie_shear_source_inversion"
            ).inversion.constant.source_galaxies.source

    phase1 = LensSourcePhase(
        phase_name="phase_1_lens_pl_shear_source_inversion",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalPowerLaw,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(
                redshift=redshift_source,
                pixelization=pl_pixelization,
                regularization=pl_regularization,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        inversion_pixel_limit=inversion_pixel_limit,
        cluster_pixel_scale=cluster_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 75
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_inversion_phase()

    return pipeline.PipelineImaging(pipeline_name, phase1)



from .phase import GalaxyFitPhase, Phase
from autolens.pipeline.phase.phase_extensions import HyperGalaxyPhase
from .phase_imaging import (
    PhaseImaging,
    MultiPlanePhase,
    LensSourcePlanePhase,
    LensPlanePhase,
)
from .phase_positions import PhasePositions



import numpy as np
import pytest
from astropy import cosmology as cosmo

import autofit as af
from autolens.lens import lens_fit
from autolens.lens import ray_tracing as rt
from autolens.model.galaxy import galaxy as g
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.hyper import hyper_data as hd
from autolens.model.inversion import pixelizations as px
from autolens.model.inversion import regularization as rg
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp
from autolens.pipeline.phase import phase_extensions
from autolens.pipeline.phase import phase_imaging
from test.unit.mock.pipeline import mock_pipeline


@pytest.fixture(name="lens_galaxy")
def make_lens_galaxy():
    return g.Galaxy(
        redshift=1.0, light=lp.SphericalSersic(), mass=mp.SphericalIsothermal()
    )


@pytest.fixture(name="source_galaxy")
def make_source_galaxy():
    return g.Galaxy(redshift=2.0, light=lp.SphericalSersic())


@pytest.fixture(name="lens_galaxies")
def make_lens_galaxies(lens_galaxy):
    lens_galaxies = af.ModelInstance()
    lens_galaxies.lens = lens_galaxy
    return lens_galaxies


@pytest.fixture(name="all_galaxies")
def make_all_galaxies(lens_galaxy, source_galaxy):
    galaxies = af.ModelInstance()
    galaxies.lens = lens_galaxy
    galaxies.source = source_galaxy
    return galaxies


@pytest.fixture(name="lens_instance")
def make_lens_instance(lens_galaxies):
    instance = af.ModelInstance()
    instance.lens_galaxies = lens_galaxies
    return instance


@pytest.fixture(name="lens_result")
def make_lens_result(lens_data_7x7, lens_instance):
    return phase_imaging.LensPlanePhase.Result(
        constant=lens_instance,
        figure_of_merit=1.0,
        previous_variable=af.ModelMapper(),
        gaussian_tuples=None,
        analysis=phase_imaging.LensPlanePhase.Analysis(
            lens_data=lens_data_7x7, cosmology=cosmo.Planck15, positions_threshold=1.0
        ),
        optimizer=None,
    )


@pytest.fixture(name="lens_source_instance")
def make_lens_source_instance(lens_galaxy, source_galaxy):
    source_galaxies = af.ModelInstance()
    lens_galaxies = af.ModelInstance()
    source_galaxies.source = source_galaxy
    lens_galaxies.lens = lens_galaxy

    instance = af.ModelInstance()
    instance.source_galaxies = source_galaxies
    instance.lens_galaxies = lens_galaxies
    return instance


@pytest.fixture(name="lens_source_result")
def make_lens_source_result(lens_data_7x7, lens_source_instance):
    return phase_imaging.LensSourcePlanePhase.Result(
        constant=lens_source_instance,
        figure_of_merit=1.0,
        previous_variable=af.ModelMapper(),
        gaussian_tuples=None,
        analysis=phase_imaging.LensSourcePlanePhase.Analysis(
            lens_data=lens_data_7x7, cosmology=cosmo.Planck15, positions_threshold=1.0
        ),
        optimizer=None,
    )


@pytest.fixture(name="multi_plane_instance")
def make_multi_plane_instance(all_galaxies):
    instance = af.ModelInstance()
    instance.galaxies = all_galaxies
    return instance


@pytest.fixture(name="multi_plane_result")
def make_multi_plane_result(lens_data_7x7, multi_plane_instance):
    return phase_imaging.MultiPlanePhase.Result(
        constant=multi_plane_instance,
        figure_of_merit=1.0,
        previous_variable=af.ModelMapper(),
        gaussian_tuples=None,
        analysis=phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_7x7, cosmology=cosmo.Planck15, positions_threshold=1.0
        ),
        optimizer=None,
    )


class MostLikelyFit(object):
    def __init__(self, model_image_2d):
        self.model_image_2d = model_image_2d


class MockResult(object):
    def __init__(self, most_likely_fit=None):
        self.most_likely_fit = most_likely_fit
        self.analysis = MockAnalysis()
        self.path_galaxy_tuples = []
        self.variable = af.ModelMapper()


class MockAnalysis(object):
    pass


class MockOptimizer(af.NonLinearOptimizer):
    def __init__(
        self,
        phase_name="mock_optimizer",
        phase_tag="tag",
        phase_folders=tuple(),
        model_mapper=None,
    ):
        super().__init__(
            phase_folders=phase_folders,
            phase_tag=phase_tag,
            phase_name=phase_name,
            model_mapper=model_mapper,
        )

    def fit(self, analysis):
        # noinspection PyTypeChecker
        return af.Result(None, analysis.fit(None), None)


class MockPhase(object):
    def __init__(self):
        self.phase_name = "phase name"
        self.phase_path = "phase_path"
        self.optimizer = MockOptimizer()
        self.phase_folders = [""]
        self.phase_tag = ""

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def run(self, *args, **kwargs):
        return MockResult()


class TestVariableFixing(object):
    def test_defaults_both(self):
        # noinspection PyTypeChecker
        phase = phase_extensions.InversionBackgroundBothPhase(MockPhase())
        mapper = af.ModelMapper()
        mapper.hyper_image_sky = hd.HyperImageSky

        prior_model = mapper.hyper_image_sky

        phase.add_defaults(mapper)

        assert isinstance(mapper.hyper_image_sky, af.PriorModel)
        assert isinstance(mapper.hyper_noise_background, af.PriorModel)

        assert mapper.hyper_image_sky.cls == hd.HyperImageSky
        assert mapper.hyper_noise_background.cls == hd.HyperNoiseBackground

        assert mapper.hyper_image_sky is prior_model

    def test_defaults_hyper_image_sky(self):
        # noinspection PyTypeChecker
        phase = phase_extensions.InversionBackgroundSkyPhase(MockPhase())

        mapper = af.ModelMapper()
        phase.add_defaults(mapper)

        assert isinstance(mapper.hyper_image_sky, af.PriorModel)
        assert mapper.hyper_image_sky.cls == hd.HyperImageSky

    def test_defaults_background_noise(self):
        # noinspection PyTypeChecker
        phase = phase_extensions.InversionBackgroundNoisePhase(MockPhase())

        mapper = af.ModelMapper()
        phase.add_defaults(mapper)

        assert isinstance(mapper.hyper_noise_background, af.PriorModel)
        assert mapper.hyper_noise_background.cls == hd.HyperNoiseBackground

    def test_make_pixelization_variable(self):
        instance = af.ModelInstance()
        mapper = af.ModelMapper()

        mapper.lens_galaxy = gm.GalaxyModel(
            redshift=g.Redshift, pixelization=px.Rectangular, regularization=rg.Constant
        )
        mapper.source_galaxy = gm.GalaxyModel(
            redshift=g.Redshift, light=lp.EllipticalLightProfile
        )

        assert mapper.prior_count == 9

        instance.lens_galaxy = g.Galaxy(
            pixelization=px.Rectangular(), regularization=rg.Constant(), redshift=1.0
        )
        instance.source_galaxy = g.Galaxy(
            redshift=1.0, light=lp.EllipticalLightProfile()
        )

        # noinspection PyTypeChecker
        phase = phase_extensions.VariableFixingHyperPhase(
            MockPhase(),
            "mock_phase",
            variable_classes=(px.Pixelization, rg.Regularization),
        )

        mapper = mapper.copy_with_fixed_priors(instance, phase.variable_classes)

        assert mapper.prior_count == 3
        assert mapper.lens_galaxy.redshift == 1.0
        assert mapper.source_galaxy.light.axis_ratio == 1.0

        phase.add_defaults(mapper)

        assert mapper.prior_count == 3
        assert mapper.lens_galaxy.redshift == 1.0
        assert mapper.source_galaxy.light.axis_ratio == 1.0


class TestImagePassing(object):
    def test_path_galaxy_tuples(self, lens_result, lens_galaxy):
        assert lens_result.path_galaxy_tuples == [
            (("lens_galaxies", "lens"), lens_galaxy)
        ]

    def test_lens_source_galaxy_dict(
        self, lens_source_result, lens_galaxy, source_galaxy
    ):
        assert lens_source_result.path_galaxy_tuples == [
            (("source_galaxies", "source"), source_galaxy),
            (("lens_galaxies", "lens"), lens_galaxy),
        ]

    def test_multi_plane_galaxy_dict(
        self, multi_plane_result, lens_galaxy, source_galaxy
    ):
        assert multi_plane_result.path_galaxy_tuples == [
            (("galaxies", "lens"), lens_galaxy),
            (("galaxies", "source"), source_galaxy),
        ]

    def test_lens_image_dict_2d_and_1d(self, lens_result, mask_7x7):
        image_2d_dict = lens_result.image_2d_dict

        assert isinstance(image_2d_dict[("lens_galaxies", "lens")], np.ndarray)
        assert image_2d_dict[("lens_galaxies", "lens")].shape == (7, 7)

        image_1d_dict = lens_result.image_1d_dict_from_mask(mask=mask_7x7)
        assert image_1d_dict[("lens_galaxies", "lens")].shape == (9,)

    def test_lens_source_image_dict(self, lens_source_result, mask_7x7):
        image_2d_dict = lens_source_result.image_2d_dict

        assert isinstance(image_2d_dict[("lens_galaxies", "lens")], np.ndarray)
        assert isinstance(image_2d_dict[("source_galaxies", "source")], np.ndarray)

        assert image_2d_dict[("lens_galaxies", "lens")].shape == (7, 7)
        assert image_2d_dict[("source_galaxies", "source")].shape == (7, 7)

        image_1d_dict = lens_source_result.image_1d_dict_from_mask(mask=mask_7x7)

        assert image_1d_dict[("lens_galaxies", "lens")].shape == (9,)
        assert image_1d_dict[("source_galaxies", "source")].shape == (9,)

        lens_source_result.constant.lens_galaxies.lens = g.Galaxy(redshift=0.5)
        lens_source_result.constant.source_galaxies.source = g.Galaxy(redshift=1.0)

    def test_multi_plane_image_dict(self, multi_plane_result):
        image_dict = multi_plane_result.image_2d_dict
        assert isinstance(image_dict[("galaxies", "lens")], np.ndarray)
        assert isinstance(image_dict[("galaxies", "source")], np.ndarray)

        multi_plane_result.constant.galaxies.lens = g.Galaxy(redshift=0.5)

        image_dict = multi_plane_result.image_2d_dict
        assert (image_dict[("galaxies", "lens")] == np.zeros((7, 7))).all()
        assert isinstance(image_dict[("galaxies", "source")], np.ndarray)

    def test_galaxy_image_dict(
        self, lens_galaxy, source_galaxy, grid_stack_7x7, convolver_image_7x7
    ):
        tracer = rt.TracerImageSourcePlanes(
            [lens_galaxy], [source_galaxy], grid_stack_7x7
        )

        assert (
            len(
                tracer.galaxy_image_dict_from_convolver_image(
                    convolver_image=convolver_image_7x7
                )
            )
            == 2
        )
        assert lens_galaxy in tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=convolver_image_7x7
        )
        assert source_galaxy in tracer.galaxy_image_dict_from_convolver_image(
            convolver_image=convolver_image_7x7
        )

    def test__results_are_passed_to_new_analysis__sets_up_hyper_images(
        self, mask_function_7x7, results_collection_7x7, ccd_data_7x7
    ):
        results_collection_7x7[0].galaxy_images = [
            2.0 * np.ones((7, 7)),
            2.0 * np.ones((7, 7)),
        ]
        results_collection_7x7[0].galaxy_images[0][3, 2] = -1.0
        results_collection_7x7[0].galaxy_images[1][3, 4] = -1.0

        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)
            ),
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_7x7,
            phase_name="test_phase",
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (
            analysis.hyper_galaxy_image_1d_path_dict[("g0",)]
            == np.array([2.0, 2.0, 2.0, 0.02, 2.0, 2.0, 2.0, 2.0, 2.0])
        ).all()

        assert (
            analysis.hyper_galaxy_image_1d_path_dict[("g1",)]
            == np.array([2.0, 2.0, 2.0, 2.0, 2.0, 0.02, 2.0, 2.0, 2.0])
        ).all()

        assert (
            analysis.hyper_model_image_1d
            == np.array([4.0, 4.0, 4.0, 2.02, 4.0, 2.02, 4.0, 4.0, 4.0])
        ).all()

    def test__results_are_passed_to_new_analysis__hyper_images_values_below_minimum_are_scaled_up_using_config(
        self, mask_function_7x7, results_collection_7x7, ccd_data_7x7
    ):
        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(
                lens=gm.GalaxyModel(redshift=0.5, hyper_galaxy=g.HyperGalaxy)
            ),
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_7x7,
            phase_name="test_phase",
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (analysis.hyper_model_image_1d == 5.0 * np.ones(9)).all()

        assert (
            analysis.hyper_galaxy_image_1d_path_dict[("g0",)] == 2.0 * np.ones(9)
        ).all()
        assert (
            analysis.hyper_galaxy_image_1d_path_dict[("g1",)] == 3.0 * np.ones(9)
        ).all()

    def test__results_are_passed_to_new_analysis__sets_up_hyper_cluster_images__includes_hyper_minimum(
        self, mask_function_7x7, results_collection_7x7, ccd_data_7x7
    ):
        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            phase_name="test_phase",
            lens_galaxies=dict(
                lens=gm.GalaxyModel(
                    redshift=0.5,
                    hyper_galaxy=g.HyperGalaxy,
                    pixelization=px.VoronoiBrightnessImage,
                    regularization=rg.Constant,
                )
            ),
            mask_function=mask_function_7x7,
            inversion_pixel_limit=5,
            cluster_pixel_scale=None,
            optimizer_class=mock_pipeline.MockNLO,
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)]
            == 2.0 * np.ones(9)
        ).all()
        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)]
            == 3.0 * np.ones(9)
        ).all()

        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(
                lens=gm.GalaxyModel(
                    redshift=0.5,
                    hyper_galaxy=g.HyperGalaxy,
                    pixelization=px.VoronoiBrightnessImage,
                    regularization=rg.Constant,
                )
            ),
            inversion_pixel_limit=1,
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_7x7,
            cluster_pixel_scale=ccd_data_7x7.pixel_scale,
            phase_name="test_phase",
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)]
            == 2.0 * np.ones(9)
        ).all()
        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)]
            == 3.0 * np.ones(9)
        ).all()
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)])
            == analysis.lens_data.cluster.shape[0]
        )
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)])
            == analysis.lens_data.cluster.shape[0]
        )

        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(
                lens=gm.GalaxyModel(
                    redshift=0.5,
                    hyper_galaxy=g.HyperGalaxy,
                    pixelization=px.VoronoiBrightnessImage,
                    regularization=rg.Constant,
                )
            ),
            inversion_pixel_limit=1,
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_7x7,
            cluster_pixel_scale=ccd_data_7x7.pixel_scale * 2.0,
            phase_name="test_phase",
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)]
            == np.array([0.5, 1.0, 1.0, 2.0])
        ).all()
        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)]
            == np.array([0.75, 1.5, 1.5, 3.0])
        ).all()
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)])
            == analysis.lens_data.cluster.shape[0]
        )
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)])
            == analysis.lens_data.cluster.shape[0]
        )

        results_collection_7x7[0].galaxy_images = [
            2.0 * np.ones((7, 7)),
            2.0 * np.ones((7, 7)),
        ]
        results_collection_7x7[0].galaxy_images[0][3, 2] = -1.0
        results_collection_7x7[0].galaxy_images[1][3, 4] = -1.0

        phase_7x7 = phase_imaging.LensSourcePlanePhase(
            lens_galaxies=dict(
                lens=gm.GalaxyModel(
                    redshift=0.5,
                    hyper_galaxy=g.HyperGalaxy,
                    pixelization=px.VoronoiBrightnessImage,
                    regularization=rg.Constant,
                )
            ),
            inversion_pixel_limit=1,
            optimizer_class=mock_pipeline.MockNLO,
            mask_function=mask_function_7x7,
            cluster_pixel_scale=ccd_data_7x7.pixel_scale * 2.0,
            phase_name="test_phase",
        )

        analysis = phase_7x7.make_analysis(
            data=ccd_data_7x7, results=results_collection_7x7
        )

        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)]
            == np.array([2.0, 2.0, 1.25, 2.0])
        ).all()
        assert (
            analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)]
            == np.array([2.0, 2.0, 2.0, 1.25])
        ).all()
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g0",)])
            == analysis.lens_data.cluster.shape[0]
        )
        assert (
            len(analysis.hyper_galaxy_cluster_image_1d_path_dict[("g1",)])
            == analysis.lens_data.cluster.shape[0]
        )

    def test_associate_images_lens(self, lens_instance, lens_result, lens_data_7x7):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", lens_result)

        analysis = phase_imaging.LensPlanePhase.Analysis(
            lens_data=lens_data_7x7,
            cosmology=None,
            positions_threshold=None,
            results=results_collection,
        )

        instance = analysis.associate_images(instance=lens_instance)

        hyper_model_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=lens_result.image_2d_dict[("lens_galaxies", "lens")]
        )

        assert instance.lens_galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )
        assert instance.lens_galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )

    def test_associate_images_lens_source(
        self, lens_source_instance, lens_source_result, lens_data_7x7
    ):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", lens_source_result)
        analysis = phase_imaging.LensSourcePlanePhase.Analysis(
            lens_data=lens_data_7x7,
            cosmology=None,
            positions_threshold=None,
            results=results_collection,
        )

        instance = analysis.associate_images(lens_source_instance)

        hyper_lens_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=lens_source_result.image_2d_dict[("lens_galaxies", "lens")]
        )
        hyper_source_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=lens_source_result.image_2d_dict[("source_galaxies", "source")]
        )

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        assert instance.lens_galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )
        assert instance.source_galaxies.source.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )

        assert instance.lens_galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_lens_image_1d, 1.0e-4
        )
        assert instance.source_galaxies.source.hyper_galaxy_image_1d == pytest.approx(
            hyper_source_image_1d, 1.04e-4
        )

    def test_associate_images_multi_plane(
        self, multi_plane_instance, multi_plane_result, lens_data_7x7
    ):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", multi_plane_result)
        analysis = phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_7x7,
            cosmology=None,
            positions_threshold=None,
            results=results_collection,
        )

        instance = analysis.associate_images(instance=multi_plane_instance)

        hyper_lens_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "lens")]
        )
        hyper_source_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "source")]
        )

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        assert instance.galaxies.lens.hyper_galaxy_image_1d == pytest.approx(
            hyper_lens_image_1d, 1.0e-4
        )
        assert instance.galaxies.source.hyper_galaxy_image_1d == pytest.approx(
            hyper_source_image_1d, 1.0e-4
        )

        assert instance.galaxies.lens.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )
        assert instance.galaxies.source.hyper_model_image_1d == pytest.approx(
            hyper_model_image_1d, 1.0e-4
        )

    def test_fit_uses_hyper_fit_correctly_multi_plane(
        self, multi_plane_instance, multi_plane_result, lens_data_7x7
    ):
        results_collection = af.ResultsCollection()
        results_collection.add("phase", multi_plane_result)
        analysis = phase_imaging.MultiPlanePhase.Analysis(
            lens_data=lens_data_7x7,
            cosmology=cosmo.Planck15,
            positions_threshold=None,
            results=results_collection,
        )

        hyper_galaxy = g.HyperGalaxy(
            contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
        )

        multi_plane_instance.galaxies.lens.hyper_galaxy = hyper_galaxy

        fit_figure_of_merit = analysis.fit(instance=multi_plane_instance)

        hyper_lens_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "lens")]
        )
        hyper_source_image_1d = lens_data_7x7.array_1d_from_array_2d(
            array_2d=multi_plane_result.image_2d_dict[("galaxies", "source")]
        )

        hyper_model_image_1d = hyper_lens_image_1d + hyper_source_image_1d

        g0 = g.Galaxy(
            redshift=0.5,
            light_profile=multi_plane_instance.galaxies.lens.light,
            mass_profile=multi_plane_instance.galaxies.lens.mass,
            hyper_galaxy=hyper_galaxy,
            hyper_model_image_1d=hyper_model_image_1d,
            hyper_galaxy_image_1d=hyper_lens_image_1d,
            hyper_minimum_value=0.0,
        )
        g1 = g.Galaxy(
            redshift=1.0, light_profile=multi_plane_instance.galaxies.source.light
        )

        tracer = rt.TracerImageSourcePlanes(
            lens_galaxies=[g0],
            source_galaxies=[g1],
            image_plane_grid_stack=lens_data_7x7.grid_stack,
        )

        fit = lens_fit.LensDataFit.for_data_and_tracer(
            lens_data=lens_data_7x7, tracer=tracer
        )

        assert (fit_figure_of_merit == fit.figure_of_merit).all()


@pytest.fixture(name="hyper_combined")
def make_combined():
    normal_phase = MockPhase()

    # noinspection PyUnusedLocal
    def run_hyper(*args, **kwargs):
        return MockResult()

    # noinspection PyTypeChecker
    hyper_combined = phase_extensions.CombinedHyperPhase(
        normal_phase,
        hyper_phase_classes=(
            phase_extensions.HyperGalaxyPhase,
            phase_extensions.InversionPhase,
        ),
    )

    for phase in hyper_combined.hyper_phases:
        phase.run_hyper = run_hyper

    return hyper_combined


class TestHyperAPI(object):
    def test_combined_result(self, hyper_combined):
        result = hyper_combined.run(None)

        assert hasattr(result, "hyper_galaxy")
        assert isinstance(result.hyper_galaxy, MockResult)

        assert hasattr(result, "inversion")
        assert isinstance(result.inversion, MockResult)

        assert hasattr(result, "hyper_combined")
        assert isinstance(result.hyper_combined, MockResult)

    def test_combine_variables(self, hyper_combined):
        result = MockResult()
        hyper_galaxy_result = MockResult()
        inversion_result = MockResult()

        hyper_galaxy_result.variable = af.ModelMapper()
        inversion_result.variable = af.ModelMapper()

        hyper_galaxy_result.variable.hyper_galaxy = g.HyperGalaxy
        hyper_galaxy_result.variable.pixelization = px.Pixelization()
        inversion_result.variable.pixelization = px.Pixelization
        inversion_result.variable.hyper_galaxy = g.HyperGalaxy()

        result.hyper_galaxy = hyper_galaxy_result
        result.inversion = inversion_result

        variable = hyper_combined.combine_variables(result)

        assert isinstance(variable.hyper_galaxy, af.PriorModel)
        assert isinstance(variable.pixelization, af.PriorModel)

        assert variable.hyper_galaxy.cls == g.HyperGalaxy
        assert variable.pixelization.cls == px.Pixelization

    def test_instantiation(self, hyper_combined):
        assert len(hyper_combined.hyper_phases) == 2

        galaxy_phase = hyper_combined.hyper_phases[0]
        pixelization_phase = hyper_combined.hyper_phases[1]

        assert galaxy_phase.hyper_name == "hyper_galaxy"
        assert isinstance(galaxy_phase, phase_extensions.HyperGalaxyPhase)

        assert pixelization_phase.hyper_name == "inversion"
        assert isinstance(pixelization_phase, phase_extensions.InversionPhase)

    def test_hyper_result(self, ccd_data_7x7):
        normal_phase = MockPhase()

        # noinspection PyTypeChecker
        phase = phase_extensions.HyperGalaxyPhase(normal_phase)

        # noinspection PyUnusedLocal
        def run_hyper(*args, **kwargs):
            return MockResult()

        phase.run_hyper = run_hyper

        result = phase.run(ccd_data_7x7)

        assert hasattr(result, "hyper_galaxy")
        assert isinstance(result.hyper_galaxy, MockResult)



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

import os

# In this pipeline, we'll perform an initialize analysis which fits an image with a source galaxy and no lens light
# component. The pipeline is as follows:

# Phase 1:

# Description: initialize the lens mass model and source light profile.
# Lens Mass: EllipitcalIsothermal + ExternalShear
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: None


def make_pipeline(
    pl_hyper_galaxies=True,
    pl_hyper_background_sky=True,
    pl_hyper_background_noise=True,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=None,
):

    ### SETUP PIPELINE AND PHASE NAMES, TAGS AND PATHS ###

    # We setup the pipeline name using the tagging module. In this case, the pipeline name is not given a tag and
    # will be the string specified below However, its good practise to use the 'tag.' function below, incase
    # a pipeline does use customized tag names.

    pipeline_name = "pipeline_init__lens_sie_shear_source_sersic"

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name=pipeline_name
    )

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit the lens galaxy's mass and one source galaxy, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.

    phase1 = phase_imaging.LensSourcePlanePhase(
        phase_name="phase_1_lens_sie_shear_source_sersic",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                mass=mp.EllipticalIsothermal,
                shear=mp.ExternalShear,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=redshift_source, light=lp.EllipticalSersic)
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 80
    phase1.optimizer.sampling_efficiency = 0.2

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=pl_hyper_galaxies,
        include_background_sky=pl_hyper_background_sky,
        include_background_noise=pl_hyper_background_noise,
    )

    return pipeline.PipelineImaging(pipeline_name, phase1)



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.plotters import plotter_util
from autolens.lens.plotters import lens_plotter_util


def plot_fit_subplot(
    fit,
    should_plot_mask=True,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image_plane_pix=True,
    units="arcsec",
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10.0,
    grid_pointsize=1,
    output_path=None,
    output_filename="sensitivity_fit",
    output_format="show",
):
    """Plot the model datas of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.sensitivity_fitting.SensitivityProfileFit
        Class containing fitting between the model datas and observed lens datas (including residual_map, chi_squared_map etc.)
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=9
    )

    mask = lens_plotter_util.get_mask(
        fit=fit.fit_normal, should_plot_mask=should_plot_mask
    )
    image_plane_pix_grid = lens_plotter_util.get_image_plane_pix_grid(
        should_plot_image_plane_pix, fit
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    kpc_per_arcsec = fit.tracer_normal.image_plane.kpc_per_arcsec

    lens_plotter_util.plot_image(
        fit=fit.fit_normal,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        positions=positions,
        grid=image_plane_pix_grid,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        position_pointsize=position_pointsize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 4)

    lens_plotter_util.plot_model_data(
        fit=fit.fit_normal,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 5)

    lens_plotter_util.plot_residual_map(
        fit=fit.fit_normal,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 6)

    lens_plotter_util.plot_chi_squared_map(
        fit=fit.fit_normal,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 7)

    lens_plotter_util.plot_model_data(
        fit=fit.fit_sensitive,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 8)

    lens_plotter_util.plot_residual_map(
        fit=fit.fit_sensitive,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plt.subplot(rows, columns, 9)

    lens_plotter_util.plot_chi_squared_map(
        fit=fit.fit_sensitive,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()



import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.pipeline.phase import phase_imaging, phase_extensions
from autolens.pipeline import pipeline
from autolens.pipeline import tagging as tag
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

# In this pipeline, we'll perform a basic analysis which initialize a lens model (the lens's light, mass and source's \
# light) and then fits the source galaxy using an inversion. This pipeline uses three phases:

# Phase 1:

# Description: initialize and subtracts the lens light model.
# Lens Light: EllipticalSersic + EllipticalExponential
# Lens Mass: None
# Source Light: None
# Previous Pipelines: None
# Prior Passing: None
# Notes: None

# Phase 2:

# Description: initialize the lens mass model and source light profile, using the lens subtracted image from phase 1.
# Lens Light: None
# Lens Mass: EllipitcalIsothermal
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: None
# Notes: Uses the lens light subtracted image from phase 1

# Phase 3:

# Description: Refine the lens light and mass models and source light profile, using priors from the previous 2 phases.
# Lens Light: EllipticalSersic + EllipticalExponential
# Lens Mass: EllipitcalIsothermal
# Source Light: EllipticalSersic
# Previous Pipelines: None
# Prior Passing: Lens light (variable -> phase 1), lens mass and source (variable -> phase 2)
# Notes: None


def make_pipeline(
    pl_align_bulge_disk_centre=False,
    pl_align_bulge_disk_axis_ratio=False,
    pl_align_bulge_disk_phi=False,
    phase_folders=None,
    tag_phases=True,
    redshift_lens=0.5,
    redshift_source=1.0,
    sub_grid_size=2,
    bin_up_factor=None,
    positions_threshold=None,
    inner_mask_radii=None,
    interp_pixel_scale=None,
):

    pipeline_name = tag.pipeline_name_from_name_and_settings(
        pipeline_name="pipeline_init__lens_sersic_exp_sie_source_sersic",
        align_bulge_disk_centre=pl_align_bulge_disk_centre,
        align_bulge_disk_axis_ratio=pl_align_bulge_disk_axis_ratio,
        align_bulge_disk_phi=pl_align_bulge_disk_phi,
    )

    phase_folders.append(pipeline_name)

    ### PHASE 1 ###

    # In phase 1, we will fit only the lens galaxy's light, where we:

    # 1) Set our priors on the lens galaxy (y,x) centre such that we assume the image is centred around the lens galaxy.
    # 2) Use a circular mask which includes the lens and source galaxy light.

    class BulgeDiskPhase(phase_imaging.LensPlanePhase):
        def pass_priors(self, results):

            if pl_align_bulge_disk_centre:
                self.lens_galaxies.lens.bulge.centre = (
                    self.lens_galaxies.lens.disk.centre
                )

            if pl_align_bulge_disk_axis_ratio:
                self.lens_galaxies.lens.bulge.axis_ratio = (
                    self.lens_galaxies.lens.disk.axis_ratio
                )

            if pl_align_bulge_disk_phi:
                self.lens_galaxies.lens.bulge.phi = self.lens_galaxies.lens.disk.phi

    phase1 = BulgeDiskPhase(
        phase_name="phase_1_lens_sersic_exp",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                bulge=lp.EllipticalSersic,
                disk=lp.EllipticalExponential,
            )
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        optimizer_class=af.MultiNest,
    )

    # You'll see these lines throughout all of the example pipelines. They are used to make MultiNest sample the \
    # non-linear parameter space faster (if you haven't already, checkout the tutorial '' in howtolens/chapter_2).

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 40
    phase1.optimizer.sampling_efficiency = 0.3

    ### PHASE 2 ###

    # In phase 2, we will fit the lens galaxy's mass and source galaxy's light, where we:

    # 1) Use a lens-subtracted image generated by subtracting model lens galaxy image from phase 1.
    # 2) Initialize the priors on the centre of the lens galaxy's mass-profile by linking them to those inferred for \
    #    the bulge of the light profile in phase 1.
    # 3) Have the option to use an annular mask removing the central light, if the inner_mask_radii parametr is input.

    class LensSubtractedPhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Light Sersic -> Sersic, Exp -> Exp ##

            self.lens_galaxies.lens.bulge = results.from_phase(
                "phase_1_lens_sersic_exp"
            ).constant.lens_galaxies.lens.bulge
            self.lens_galaxies.lens.disk = results.from_phase(
                "phase_1_lens_sersic_exp"
            ).constant.lens_galaxies.lens.disk

            ## Lens Mass, Move centre priors to centre of lens light ###

            self.lens_galaxies.lens.mass.centre = (
                results.from_phase("phase_1_lens_sersic_exp")
                .variable_absolute(a=0.1)
                .lens_galaxies.lens.bulge.centre
            )

    phase2 = LensSubtractedPhase(
        phase_name="phase_2_lens_sie_source_sersic",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                bulge=lp.EllipticalSersic,
                disk=lp.EllipticalExponential,
                mass=mp.EllipticalIsothermal,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=redshift_source, light=lp.EllipticalSersic)
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        inner_mask_radii=inner_mask_radii,
        interp_pixel_scale=interp_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase2.optimizer.const_efficiency_mode = False
    phase2.optimizer.n_live_points = 50
    phase2.optimizer.sampling_efficiency = 0.3

    ### PHASE 3 ###

    # In phase 3, we will fit simultaneously the lens and source galaxies, where we:

    # 1) Initialize the lens's light, mass, shear and source's light using the results of phases 1 and 2.
    # 2) Use a circular mask, to fully capture the lens and source light.

    class LensSourcePhase(phase_imaging.LensSourcePlanePhase):
        def pass_priors(self, results):

            ## Lens Light, Sersic -> Sersic ###

            self.lens_galaxies.lens.bulge = results.from_phase(
                "phase_1_lens_sersic_exp"
            ).variable.lens_galaxies.lens.bulge

            self.lens_galaxies.lens.disk = results.from_phase(
                "phase_1_lens_sersic_exp"
            ).variable.lens_galaxies.lens.disk

            ## Lens Mass, SIE -> SIE, Shear -> Shear ###

            self.lens_galaxies.lens.mass = results.from_phase(
                "phase_2_lens_sie_source_sersic"
            ).variable.lens_galaxies.lens.mass

            ### Source Light, Sersic -> Sersic, Exp -> Exp ###

            self.source_galaxies.source = results.from_phase(
                "phase_2_lens_sie_source_sersic"
            ).variable.source_galaxies.source

    phase3 = LensSourcePhase(
        phase_name="phase_3_lens_sersic_exp_sie_source_sersic",
        phase_folders=phase_folders,
        tag_phases=tag_phases,
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=redshift_lens,
                bulge=lp.EllipticalSersic,
                disk=lp.EllipticalExponential,
                mass=mp.EllipticalIsothermal,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=redshift_source, light=lp.EllipticalSersic)
        ),
        sub_grid_size=sub_grid_size,
        bin_up_factor=bin_up_factor,
        positions_threshold=positions_threshold,
        interp_pixel_scale=interp_pixel_scale,
        optimizer_class=af.MultiNest,
    )

    phase3.optimizer.const_efficiency_mode = True
    phase3.optimizer.n_live_points = 75
    phase3.optimizer.sampling_efficiency = 0.3

    return pipeline.PipelineImaging(pipeline_name, phase1, phase2, phase3)



import os

import pytest

from autolens.lens.plotters import sensitivity_fit_plotters


@pytest.fixture(name="sensitivity_fit_plotter_path")
def make_sensitivity_fit_plotter_setup():
    return "{}/../../test_files/plotting/fit/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__fit_sub_plot__output_dependent_on_config(
    sensitivity_fit_7x7, sensitivity_fit_plotter_path, plot_patch
):
    sensitivity_fit_plotters.plot_fit_subplot(
        fit=sensitivity_fit_7x7,
        should_plot_mask=True,
        extract_array_from_mask=True,
        zoom_around_mask=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=sensitivity_fit_plotter_path,
        output_format="png",
    )

    assert sensitivity_fit_plotter_path + "sensitivity_fit.png" in plot_patch.paths



import numpy as np
import scipy.signal

from autolens.data import ccd
from autolens.data.array import scaled_array


class MockImage(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockBackgroundNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockPoissonNoiseMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockExposureTimeMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockBackgrondSkyMap(object):
    def __new__(cls, shape, value, pixel_scale=1.0):

        return scaled_array.ScaledSquarePixelArray(
            array=value * np.ones(shape=shape), pixel_scale=pixel_scale
        )


class MockPSF(object):
    def __new__(cls, shape, value, pixel_scale=1.0, *args, **kwargs):

        return ccd.PSF(
            array=value * np.ones(shape=shape),
            pixel_scale=pixel_scale,
            origin=(0.0, 0.0),
        )


class MockImage1D(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):

        array = value * np.ones(shape=shape)

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockNoiseMap1D(np.ndarray):
    def __new__(cls, shape, value, pixel_scale=1.0):

        array = value * np.ones(shape=shape)

        obj = np.array(array, dtype="float64").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj


class MockCCDData(ccd.CCDData):
    def __init__(
        self,
        image,
        pixel_scale,
        psf,
        noise_map,
        background_noise_map,
        poisson_noise_map,
        exposure_time_map,
        background_sky_map,
        name,
    ):

        super(MockCCDData, self).__init__(
            image=image,
            pixel_scale=pixel_scale,
            psf=psf,
            noise_map=noise_map,
            background_noise_map=background_noise_map,
            poisson_noise_map=poisson_noise_map,
            exposure_time_map=exposure_time_map,
            background_sky_map=background_sky_map,
            name=name,
        )



import os

import autofit as af
from autolens.model.galaxy import galaxy_model as gm
from autolens.model.profiles import light_profiles as lp, mass_profiles as mp
from autolens.pipeline.phase import phase_imaging
from autolens.pipeline import pipeline as pl
from test.integration import integration_util
from test.simulation import simulation_util

test_type = "lens_and_source"
test_name = "lens_light_and_mass_x1_source_x1"

test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
output_path = test_path + "output/"
config_path = test_path + "config"
af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)


def pipeline():
    integration_util.reset_paths(test_name=test_name, output_path=output_path)
    ccd_data = simulation_util.load_test_ccd_data(
        data_type="lens_and_source_smooth", data_resolution="LSST"
    )
    pipeline = make_pipeline(test_name=test_name)
    pipeline.run(data=ccd_data, data_name=test_name)


def make_pipeline(test_name):

    phase1 = phase_imaging.LensSourcePlanePhase(
        phase_name="phase_1",
        phase_folders=[test_type, test_name],
        lens_galaxies=dict(
            lens=gm.GalaxyModel(
                redshift=0.5,
                light=lp.SphericalDevVaucouleurs,
                mass=mp.EllipticalIsothermal,
            )
        ),
        source_galaxies=dict(
            source=gm.GalaxyModel(redshift=1.0, light=lp.EllipticalSersic)
        ),
        optimizer_class=af.MultiNest,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.8

    return pl.PipelineImaging(test_name, phase1)


if __name__ == "__main__":
    pipeline()
